@@ -1,8 +1,9 @@
 import random
 import warnings
 import copy
-from ctypes import CDLL, c_double
+from ctypes import CDLL, POINTER, c_double, c_char_p, pointer
 from ctypes.util import find_library
+import fnmatch
 
 import redis
 import redis.client
@@ -12,11 +13,12 @@ DATABASES = {}
 
 _libc = CDLL(find_library('c'))
 _libc.strtod.restype = c_double
+_libc.strtod.argtypes = [c_char_p, POINTER(c_char_p)]
 _strtod = _libc.strtod
 
 
-class FakeRedis(object):
-    def __init__(self, db=0):
+class FakeStrictRedis(object):
+    def __init__(self, db=0, **kwargs):
         if db not in DATABASES:
             DATABASES[db] = {}
         self._db = DATABASES[db]
@@ -95,8 +97,9 @@ class FakeRedis(object):
                                       "range.")
         return self._db[name]
 
-    def keys(self):
-        return self._db.keys()
+    def keys(self, pattern=None):
+        return [key for key in self._db.keys()
+                if not key or not pattern or fnmatch.fnmatch(key, pattern)]
 
     def mget(self, keys, *args):
         all_keys = self._list_or_args(keys, args)
@@ -169,7 +172,7 @@ class FakeRedis(object):
         self._db[name] = ''.join(reconstructed)
 
     def setex(self, name, time, value):
-        pass
+        return self.set(name, value)
 
     def setnx(self, name, value):
         if name in self._db:
@@ -223,7 +226,7 @@ class FakeRedis(object):
         return any_deleted
 
     def sort(self, name, start=None, num=None, by=None, get=None, desc=False,
-             alpha=False, store=None) :
+             alpha=False, store=None):
         """Sort and return the list, set or sorted set at ``name``.
 
         ``start`` and ``num`` allow for paging through the sorted data
@@ -258,7 +261,7 @@ class FakeRedis(object):
             else:
                 data.sort()
             if not (start is None and num is None):
-                data = data[start:start+num]
+                data = data[start:start + num]
             if desc:
                 data = list(reversed(data))
             if store is not None:
@@ -298,7 +301,15 @@ class FakeRedis(object):
     def _strtod_key_func(self, arg):
         # str()'ing the arg is important! Don't ever remove this.
         arg = str(arg)
-        return _strtod(arg, None)
+        end = c_char_p()
+        val = _strtod(arg, pointer(end))
+        # real Redis also does an isnan check, not sure if
+        # that's needed here or not.
+        if end.value:
+            raise redis.ResponseError(
+                "One or more scores can't be converted into double")
+        else:
+            return val
 
     def _sort_using_by_arg(self, data, by):
         def _by_key(arg):
@@ -310,8 +321,8 @@ class FakeRedis(object):
                 return self._db.get(key)
         data.sort(key=_by_key)
 
-    def lpush(self, name, value):
-        self._db.setdefault(name, []).insert(0, value)
+    def lpush(self, name, *values):
+        self._db.setdefault(name, [])[0:0] = list(reversed((values)))
         return len(self._db[name])
 
     def lrange(self, name, start, end):
@@ -324,7 +335,7 @@ class FakeRedis(object):
     def llen(self, name):
         return len(self._db.get(name, []))
 
-    def lrem(self, name, value, count=0):
+    def lrem(self, name, count, value):
         a_list = self._db.get(name, [])
         found = []
         for i, el in enumerate(a_list):
@@ -342,8 +353,8 @@ class FakeRedis(object):
             del a_list[index]
         return len(indices_to_remove)
 
-    def rpush(self, name, value):
-        self._db.setdefault(name, []).append(value)
+    def rpush(self, name, *values):
+        self._db.setdefault(name, []).extend(list(values))
         return len(self._db[name])
 
     def lpop(self, name):
@@ -368,12 +379,13 @@ class FakeRedis(object):
         try:
             val = self._db[name]
         except KeyError:
-            return
+            return True
         if end == -1:
             end = None
         else:
             end += 1
         self._db[name] = val[start:end]
+        return True
 
     def lindex(self, name, index):
         try:
@@ -437,12 +449,14 @@ class FakeRedis(object):
             self._db[dst] = [el]
         return el
 
-    def hdel(self, name, key):
-        try:
-            del self._db.get(name, {})[key]
-            return True
-        except KeyError:
-            return False
+    def hdel(self, name, *keys):
+        h = self._db.get(name, {})
+        rem = 0
+        for k in keys:
+            if k in h:
+                del h[k]
+                rem += 1
+        return rem > 0
 
     def hexists(self, name, key):
         "Returns a boolean indicating if ``key`` exists within hash ``name``"
@@ -478,8 +492,9 @@ class FakeRedis(object):
         Set ``key`` to ``value`` within hash ``name``
         Returns 1 if HSET created a new field, otherwise 0
         """
+        key_is_new = key not in self._db.get(name, {})
         self._db.setdefault(name, {})[key] = value
-        return 1
+        return 1 if key_is_new else 0
 
     def hsetnx(self, name, key, value):
         """
@@ -510,14 +525,12 @@ class FakeRedis(object):
         "Return the list of values within hash ``name``"
         return self._db.get(name, {}).values()
 
-    def sadd(self, name, value):
+    def sadd(self, name, *values):
         "Add ``value`` to set ``name``"
         a_set = self._db.setdefault(name, set())
-        if value in a_set:
-            return False
-        else:
-            a_set.add(value)
-            return True
+        card = len(a_set)
+        a_set |= set(values)
+        return len(a_set) - card
 
     def scard(self, name):
         "Return the number of elements in set ``name``"
@@ -563,7 +576,7 @@ class FakeRedis(object):
 
     def smembers(self, name):
         "Return all members of the set ``name``"
-        return self._db.get(name)
+        return self._db.get(name, set())
 
     def smove(self, src, dst, value):
         try:
@@ -587,13 +600,12 @@ class FakeRedis(object):
             index = random.randint(0, len(members) - 1)
             return list(members)[index]
 
-    def srem(self, name, value):
+    def srem(self, name, *values):
         "Remove ``value`` from set ``name``"
-        try:
-            self._db.get(name, set()).remove(value)
-            return True
-        except KeyError:
-            return False
+        a_set = self._db.setdefault(name, set())
+        card = len(a_set)
+        a_set -= set(values)
+        return (card - len(a_set)) > 0
 
     def sunion(self, keys, *args):
         "Return the union of sets specifiued by ``keys``"
@@ -612,24 +624,37 @@ class FakeRedis(object):
         self._db[dest] = union
         return len(union)
 
-    def zadd(self, name, value=None, score=None, **pairs):
+    def zadd(self, name, *args, **kwargs):
         """
-        For each kwarg in ``pairs``, add that item and it's score to the
-        sorted set ``name``.
+        Set any number of score, element-name pairs to the key ``name``. Pairs
+        can be specified in two ways:
 
-        The ``value`` and ``score`` arguments are deprecated.
+        As *args, in the form of: score1, name1, score2, name2, ...
+        or as **kwargs, in the form of: name1=score1, name2=score2, ...
+
+        The following example would add four values to the 'my-key' key:
+        redis.zadd('my-key', 1.1, 'name1', 2.2, 'name2', name3=3.3, name4=4.4)
         """
-        if value is not None or score is not None:
-            if value is None or score is None:
-                raise redis.RedisError(
-                    "Both 'value' and 'score' must be specified to ZADD")
-            warnings.warn(DeprecationWarning(
-                "Passing 'value' and 'score' has been deprecated. " \
-                "Please pass via kwargs instead."))
-        else:
-            value = pairs.keys()[0]
-            score = pairs.values()[0]
-        self._db.setdefault(name, {})[value] = score
+        if len(args) % 2 != 0:
+            raise redis.RedisError("ZADD requires an equal number of "
+                                   "values and scores")
+        zset = self._db.setdefault(name, {})
+        added = 0
+        for score, value in zip(*[args[i::2] for i in range(2)]):
+            if value not in zset:
+                added += 1
+            try:
+                zset[value] = float(score)
+            except ValueError:
+                raise redis.ResponseError("value is not a valid float")
+        for value, score in kwargs.items():
+            if value not in zset:
+                added += 1
+            try:
+                zset[value] = float(score)
+            except ValueError:
+                raise redis.ResponseError("value is not a valid float")
+        return added
 
     def zcard(self, name):
         "Return the number of elements in the sorted set ``name``"
@@ -702,7 +727,7 @@ class FakeRedis(object):
         return [el[0] for el in in_order]
 
     def zrangebyscore(self, name, min, max,
-            start=None, num=None, withscores=False):
+                      start=None, num=None, withscores=False):
         """
         Return a range of values from the sorted set ``name`` with scores
         between ``min`` and ``max``.
@@ -728,7 +753,7 @@ class FakeRedis(object):
             if min <= all_items[item] <= max:
                 matches.append(item)
         if start is not None:
-            matches = matches[start:start+num]
+            matches = matches[start:start + num]
         if withscores:
             return [(k, all_items[k]) for k in matches]
         return matches
@@ -745,13 +770,15 @@ class FakeRedis(object):
         except ValueError:
             return None
 
-    def zrem(self, name, value):
+    def zrem(self, name, *values):
         "Remove member ``value`` from sorted set ``name``"
-        try:
-            del self._db[name][value]
-            return True
-        except KeyError:
-            return False
+        z = self._db.get(name, {})
+        rem = 0
+        for v in values:
+            if v in z:
+                del z[v]
+                rem += 1
+        return rem > 0
 
     def zremrangebyrank(self, name, min, max):
         """
@@ -798,7 +825,7 @@ class FakeRedis(object):
         return self.zrange(name, start, num, True, withscores)
 
     def zrevrangebyscore(self, name, max, min,
-            start=None, num=None, withscores=False):
+                         start=None, num=None, withscores=False):
         """
         Return a range of values from the sorted set ``name`` with scores
         between ``min`` and ``max`` in descending order.
@@ -866,15 +893,11 @@ class FakeRedis(object):
         self._db[dest] = new_zset
 
     def _list_or_args(self, keys, args):
-        # Taken directly from redis-py.
+        # Based off of list_or_args from redis-py.
         # Returns a single list combining keys and args.
-        try:
-            iter(keys)
-            # a string can be iterated, but indicates
-            # keys wasn't passed as a list
-            if isinstance(keys, basestring):
-                keys = [keys]
-        except TypeError:
+        # A string can be iterated, but indicates
+        # keys wasn't passed as a list.
+        if isinstance(keys, basestring):
             keys = [keys]
         if args:
             keys.extend(args)
@@ -904,8 +927,35 @@ class FakeRedis(object):
         raise redis.WatchError('Could not run transaction after 5 tries')
 
 
+class FakeRedis(FakeStrictRedis):
+    def setex(self, name, value, time):
+        return super(FakeRedis, self).setex(name, time, value)
+
+    def lrem(self, name, value, num=0):
+        return super(FakeRedis, self).lrem(name, num, value)
+
+    def zadd(self, name, value=None, score=None, **pairs):
+        """
+        For each kwarg in ``pairs``, add that item and it's score to the
+        sorted set ``name``.
+
+        The ``value`` and ``score`` arguments are deprecated.
+        """
+        if value is not None or score is not None:
+            if value is None or score is None:
+                raise redis.RedisError(
+                    "Both 'value' and 'score' must be specified to ZADD")
+            warnings.warn(DeprecationWarning(
+                "Passing 'value' and 'score' has been deprecated. "
+                "Please pass via kwargs instead."))
+        else:
+            value = pairs.keys()[0]
+            score = pairs.values()[0]
+        self._db.setdefault(name, {})[value] = score
+
+
 class FakePipeline(object):
-    """Helper class for FakeRedis to implement pipelines.
+    """Helper class for FakeStrictRedis to implement pipelines.
 
     A pipeline is a collection of commands that
     are buffered until you call ``execute``, at which
@@ -914,10 +964,10 @@ class FakePipeline(object):
 
     """
     def __init__(self, owner, transaction=True):
-        """Create a pipeline for the specified FakeRedis instance.
+        """Create a pipeline for the specified FakeStrictRedis instance.
 
         Arguments --
-            owner -- a FakeRedis instance.
+            owner -- a FakeStrictRedis instance.
 
         """
         self.owner = owner
@@ -928,20 +978,22 @@ class FakePipeline(object):
         self.watching = {}
 
     def __getattr__(self, name):
-        """Magic method to allow FakeRedis commands to be called.
+        """Magic method to allow FakeStrictRedis commands to be called.
 
         Returns a method that records the command for later.
 
         """
         if not hasattr(self.owner, name):
-            raise AttributeError('%r: does not have attribute %r' % (self.owner,
-                                                                     name))
+            raise AttributeError('%r: does not have attribute %r' %
+                                 (self.owner, name))
+
         def meth(*args, **kwargs):
             if self.is_immediate:
                 # Special mode during watch_multi sequence.
                 return getattr(self.owner, name)(*args, **kwargs)
             self.commands.append((name, args, kwargs))
             return self
+
         setattr(self, name, meth)
         return meth
 
@@ -956,18 +1008,24 @@ class FakePipeline(object):
         if self.watching:
             mismatches = [
                 (k, v, u) for (k, v, u) in
-                [(k, v, self.owner._db[k]) for (k, v) in self.watching.items()]
+                [(k, v, self.owner._db.get(k))
+                    for (k, v) in self.watching.items()]
                 if v != u]
             if mismatches:
                 self.commands = []
-                raise redis.WatchError('Watched key%s %s changed'
-                    % ('' if len(mismatches) == 1 else 's', ', '.join(
-                        k for (k, _, _) in mismatches)))
-        return [getattr(self.owner, name)(*args, **kwargs)
-                for name, args, kwargs in self.commands]
+                self.watching = {}
+                raise redis.WatchError(
+                    'Watched key%s %s changed' % (
+                        '' if len(mismatches) == 1 else
+                        's', ', '.join(k for (k, _, _) in mismatches)))
+        ret = [getattr(self.owner, name)(*args, **kwargs)
+               for name, args, kwargs in self.commands]
+        self.commands = []
+        self.watching = {}
+        return ret
 
     def watch(self, *keys):
-        self.watching.update((key, copy.deepcopy(self.owner._db[key]))
+        self.watching.update((key, copy.deepcopy(self.owner._db.get(key)))
                              for key in keys)
         self.need_reset = True
         self.is_immediate = True
